import logging
import re
import json
import random
import pkg_resources

from collections import namedtuple
from typing import List, Generator

import er
from typeguard import typechecked

logger = logging.getLogger(__name__)


class InvalidRegex(Exception):
    pass


dictionary_password = namedtuple("dictionary_password", ["display", "password"])


@typechecked
def get_word_length_combinations(
    *, valid_word_lengths: List[int], character_target: int, word_count: int = 3, current_numbers=[]
) -> Generator[list, None, None]:
    current_sum = sum(current_numbers)
    if len(current_numbers) > word_count:
        return
    if current_sum == character_target and len(current_numbers) == word_count:
        yield current_numbers
    if current_sum >= character_target:
        return
    for number in valid_word_lengths:
        yield from get_word_length_combinations(
            valid_word_lengths=valid_word_lengths,
            character_target=character_target,
            current_numbers=current_numbers + [number],
            word_count=word_count,
        )


@typechecked
def password_from_words(*, length: int = 15, word_count: int = 3, valid_word_lengths: List[int] = list(range(3, 9))):
    logger.debug(
        f"creating english word password with length {length}, "
        f"word_count {word_count}, and valid_word_lengths {sorted(valid_word_lengths)}."
    )

    if length not in list(range(1, 29)):
        raise ValueError("Password length must be 1 - 28.")
    if word_count not in list(range(1, 11)):
        raise ValueError("Word count must be 1 - 10.")

    with open(pkg_resources.resource_filename("ppa_tools", "config/words.json"), "r") as open_file:
        words = json.loads("\n".join(open_file.readlines()))

    if all([length > 20, word_count <= 2]):
        # The longest word is 10 characters to make them easier to remember.
        raise ValueError("Cannot create a password longer than 20 characters using no more than 2 words.")

    # Randomise the list of ints before feeding them to the algorithm, making the
    # generator return a different order of results every time. This means we can
    # take the first value from the generator and still have it be different every time.
    random.shuffle(valid_word_lengths)

    try:
        word_lengths = next(
            get_word_length_combinations(
                valid_word_lengths=valid_word_lengths, character_target=length, word_count=word_count
            )
        )
        password = [random.choice(words[str(word_length)]) for word_length in word_lengths]
        random.shuffle(password)
        return dictionary_password(display=" ".join(password), password="".join(password))
    except (StopIteration, IndexError):
        raise ValueError(
            "The supplied password configuration is either not compatible "
            "with the word list, or describes an impossible password."
        )


@typechecked
def password_from_regex(regex: str = r"[A-Za-z0-9\-£\:\$\#\%\&\^\\]", length: int = 24, shuffle: bool = True) -> str:
    try:
        return er.generate(f"{regex}{{{length}}}", shuffle=shuffle)
    except re.error as e:
        raise InvalidRegex(f"Failed to compile supplied regex {regex}: {str(e)}")
