import logging

from . import _client, exceptions

logger = logging.getLogger(__name__)


def _get_element_choices(service_now, *, table: str, element: str) -> dict:
    """
    Example full output from ServiceNow. We only want the label and value
    keys from each dictionary in the result list.

    {
      "result": [
        {
          "dependent_value": "",
          "sys_mod_count": "0",
          "language": "en",
          "label": "1 - Critical",
          "sys_updated_on": "2019-09-02 11:22:28",
          "sys_domain_path": "/",
          "sys_tags": "",
          "sequence": "1",
          "sys_id": "a5a5f6871b233300dc0331d7dc4bcb0d",
          "inactive": "false",
          "sys_updated_by": "system",
          "sys_created_on": "2019-09-02 11:22:28",
          "hint": "",
          "sys_domain": {
            "link": "https://dev89629.service-now.com/api/now/v1/table/sys_user_group/global",
            "value": "global"
          },
          "name": "task",
          "value": "1",
          "sys_created_by": "system",
          "element": "priority"
        }
      ]
    }
    """
    response = _client.get(service_now, "table/sys_choice", params={"name": table, "element": element})

    try:
        results = response["result"]
    except KeyError:
        logger.error(response)
        raise exceptions.ServiceNowRequestError(
            (
                "No result set found in response when getting valid element choices. "
                "Please view the logs for more information."
            )
        )

    if not results:
        raise exceptions.NoSuchObject(
            f"Empty result set returned from '{table}' table when getting choices for element '{element}'."
        )

    choices = {}
    for result in results:
        label = result.get("label")
        value = result.get("value")
        if any([label is None, value is None]):
            logger.error(
                f"got unusable choice result for element '{element}' in table "
                f"'{table}', label: '{label}', value: '{value}"
            )
        else:
            choices[label] = value

    return choices
