from collections import namedtuple
from typing import NewType
import logging

logger = logging.getLogger(__name__)

user_fields = sorted(
    [
        "calendar_integration",
        "country",
        "user_password",
        "last_login_time",
        "source",
        "sys_updated_on",
        "building",
        "web_service_access_only",
        "notification",
        "enable_multifactor_authn",
        "sys_updated_by",
        "sys_created_on",
        "sys_domain",
        "state",
        "vip",
        "sys_created_by",
        "zip",
        "home_phone",
        "time_format",
        "last_login",
        "default_perspective",
        "active",
        "sys_domain_path",
        "cost_center",
        "phone",
        "name",
        "employee_number",
        "password_needs_reset",
        "gender",
        "city",
        "failed_attempts",
        "user_name",
        "roles",
        "title",
        "sys_class_name",
        "sys_id",
        "internal_integration_user",
        "ldap_server",
        "mobile_phone",
        "street",
        "company",
        "department",
        "first_name",
        "email",
        "introduction",
        "preferred_language",
        "manager",
        "locked_out",
        "sys_mod_count",
        "last_name",
        "photo",
        "middle_name",
        "sys_tags",
        "time_zone",
        "schedule",
        "date_format",
        "location",
    ]
)
# Silence the following mypy error 'error: List or tuple literal expected as the second argument to namedtuple()'
_user_record = namedtuple("User", user_fields)  # type: ignore
User = NewType("User", _user_record)
requested_item_fields = sorted(
    [
        "parent",
        "made_sla",
        "watch_list",
        "sc_catalog",
        "upon_reject",
        "sys_updated_on",
        "approval_history",
        "number",
        "sys_updated_by",
        "opened_by",
        "user_input",
        "price",
        "sys_created_on",
        "recurring_frequency",
        "sys_domain",
        "context",
        "state",
        "sys_created_by",
        "knowledge",
        "order",
        "closed_at",
        "cmdb_ci",
        "delivery_plan",
        "backordered",
        "impact",
        "active",
        "work_notes_list",
        "business_service",
        "priority",
        "sys_domain_path",
        "time_worked",
        "expected_start",
        "opened_at",
        "business_duration",
        "group_list",
        "configuration_item",
        "work_end",
        "approval_set",
        "work_notes",
        "order_guide",
        "request",
        "short_description",
        "correlation_display",
        "delivery_task",
        "work_start",
        "assignment_group",
        "additional_assignee_list",
        "description",
        "calendar_duration",
        "close_notes",
        "sys_class_name",
        "closed_by",
        "follow_up",
        "sys_id",
        "contact_type",
        "urgency",
        "company",
        "reassignment_count",
        "activity_due",
        "assigned_to",
        "comments",
        "quantity",
        "approval",
        "sla_due",
        "comments_and_work_notes",
        "due_date",
        "sys_mod_count",
        "recurring_price",
        "sys_tags",
        "billable",
        "cat_item",
        "stage",
        "escalation",
        "upon_approval",
        "correlation_id",
        "location",
        "estimated_delivery",
    ]
)
# Silence the following mypy error 'error: List or tuple literal expected as the second argument to namedtuple()'
_requested_item_record = namedtuple("RequestedItem", requested_item_fields)  # type: ignore
RequestedItem = NewType("RequestedItem", _requested_item_record)
incident_fields = sorted(
    [
        "parent",
        "made_sla",
        "caused_by",
        "watch_list",
        "upon_reject",
        "sys_updated_on",
        "child_incidents",
        "hold_reason",
        "approval_history",
        "number",
        "resolved_by",
        "sys_updated_by",
        "opened_by",
        "user_input",
        "sys_created_on",
        "sys_domain",
        "state",
        "sys_created_by",
        "knowledge",
        "order",
        "calendar_stc",
        "closed_at",
        "cmdb_ci",
        "delivery_plan",
        "impact",
        "active",
        "work_notes_list",
        "business_service",
        "priority",
        "sys_domain_path",
        "rfc",
        "time_worked",
        "expected_start",
        "opened_at",
        "business_duration",
        "group_list",
        "work_end",
        "caller_id",
        "reopened_time",
        "resolved_at",
        "approval_set",
        "subcategory",
        "work_notes",
        "short_description",
        "close_code",
        "correlation_display",
        "delivery_task",
        "work_start",
        "assignment_group",
        "additional_assignee_list",
        "business_stc",
        "description",
        "calendar_duration",
        "close_notes",
        "notify",
        "sys_class_name",
        "closed_by",
        "follow_up",
        "parent_incident",
        "sys_id",
        "contact_type",
        "reopened_by",
        "incident_state",
        "urgency",
        "problem_id",
        "company",
        "reassignment_count",
        "activity_due",
        "assigned_to",
        "severity",
        "comments",
        "approval",
        "sla_due",
        "comments_and_work_notes",
        "due_date",
        "sys_mod_count",
        "reopen_count",
        "sys_tags",
        "escalation",
        "upon_approval",
        "correlation_id",
        "location",
        "category",
    ]
)
# Silence the following mypy error 'error: List or tuple literal expected as the second argument to namedtuple()'
_incident_record = namedtuple("Incident", incident_fields)  # type: ignore
Incident = NewType("Incident", _incident_record)


def _generate_named_tuples(name, fields):
    def wrapper(func):
        def _fn(connection, *args, **kwargs):
            # If return attributes are supplied, get them from kwargs
            # and use them when creating the new records.
            attributes = kwargs.get("attributes") if kwargs.get("attributes") else fields
            response = func(connection, *args, **kwargs)
            try:
                records = response["result"]
            except KeyError:
                # No records were returned. By this point we know that the status code
                # was deemed acceptable, and that the JSON response parsing worked. Those
                # are handled by the @request decorator in the _client module. It is safe
                # to return an empty list here, as the public API functions will raise
                # if they need to.
                return []

            if isinstance(records, dict):
                records = [records]

            record_tuple = namedtuple(name, attributes)  # type: ignore
            record_data = []
            for record in records:
                current_record = []
                for field in attributes:
                    field_data = record.get(field)
                    if isinstance(field_data, dict):
                        # Silence the following mypy error
                        # 'error: namedtuple() expects a string literal as the first argument'
                        inner_record = namedtuple(field, field_data.keys())  # type: ignore
                        # Silence the following mypy error 'error: Too many arguments for named tuple'
                        current_record.append(inner_record(*field_data.values()))  # type: ignore

                    else:
                        current_record.append(field_data)

                # Silence the following mypy error 'error: Too many arguments for named tuple'
                record_data.append(record_tuple(*current_record))  # type: ignore
            return record_data

        return _fn

    return wrapper
