import logging

from typing import List, Optional

from typeguard import typechecked

from ._client import ServiceNow
from . import _incidents, _config_reader, exceptions
from ._records import Incident


logger = logging.getLogger(__name__)


@typechecked
def get_all(
    service_now: ServiceNow, *, display_value: Optional[str] = "false", attributes: Optional[list] = None
) -> List[Incident]:
    """
    Get all ServiceNow Incidents.

    :param service_now: A ServiceNow Adaptor
    :param attributes: A list of attributes to return in each Incident Record.
    :param display_value: Return field display values (true), actual values (false), or both (all).

    :raises exceptions.AuthenticationFailed: if authentication failed using the supplied access token.

    :returns: A list of :ref:`records:Incident` records.
    """
    return _incidents.get_all(service_now, display_value=display_value, attributes=attributes)


@typechecked
def get_by_sys_id(
    service_now: ServiceNow, sys_id: str, *, display_value: Optional[str] = "false", attributes: Optional[list] = None
) -> Incident:
    """
    Get a ServiceNow incident by sys_id.

    :param service_now: A ServiceNow Adaptor
    :param sys_id: The sys id of the ServiceNow incident.
    :param attributes: A list of attributes to return in the Incident Record.
    :param display_value: Return field display values (true), actual values (false), or both (all).

    :raises exceptions.AuthenticationFailed: if authentication failed using the supplied access token.
    :raises exceptions.NoSuchObject: if no matching incident was found.

    :returns: A :ref:`records:Incident` record.
    """
    logger.info(f"looking for incident with sys_id {sys_id}")
    with exceptions.raise_on_empty(f"No incident found with sys_id {sys_id}"):
        return _incidents.by_sys_id(service_now, sys_id, display_value=display_value, attributes=attributes)[0]


@typechecked
def get_by_number(
    service_now: ServiceNow, number: str, *, display_value: Optional[str] = "false", attributes: Optional[list] = None
) -> List[Incident]:
    """
    Get a ServiceNow incident with a specific number.

    :param service_now: A ServiceNow Adaptor
    :param number: The incident number to look for.
    :param attributes: A list of attributes to return in the Incident Record.
    :param display_value: Return field display values (true), actual values (false), or both (all).

    :raises exceptions.AuthenticationFailed: if authentication failed using the supplied access token.
    :raises exceptions.NoSuchObject: if no matching incident was found.

    :returns: A list of :ref:`records:Incident` records.
    """
    logger.info(f"looking for incident with number {number}")
    with exceptions.raise_on_empty(f"No incident found with number {number}"):
        results = _incidents.search(service_now, display_value=display_value, attributes=attributes, number=number)
        if not results:
            # Caught by raise_on_empty
            raise IndexError
    return results


@typechecked
def create(service_now: ServiceNow, display_value: Optional[str] = "false", **incident_data) -> Incident:
    """
    Create a ServiceNow incident.

    :param service_now: A ServiceNow Adaptor.
    :param incident_data: Keyword arguments to use when creating an Incident. Each keyword must
        be a valid Incident field.
    :param display_value: Return field display values (true), actual values (false), or both (all).

    :raises exceptions.NoSuchObject: if no updated Incident record was returned.

    :returns: A :ref:`records:Incident` record.
    """
    logger.info(f"creating incident with parameters {incident_data}")
    with exceptions.raise_on_empty("No new incident was returned. Please view the logs for more information."):
        return _incidents.create(service_now, display_value=display_value, **incident_data)[0]


@typechecked
def update(service_now: ServiceNow, sys_id: str, display_value: Optional[str] = "false", **update_params) -> Incident:
    """
    Create a ServiceNow incident.

    :param service_now: A ServiceNow Adaptor.
    :param update_params: Keyword arguments to use when updating an Incident. Each keyword must
        be a valid Incident field.
    :param display_value: Return field display values (true), actual values (false), or both (all).

    :raises exceptions.NoSuchObject: if no updated Incident record was returned.

    :returns: A :ref:`records:Incident` record.
    """
    logger.info(f"updating incident with sys_id {sys_id} and parameters {update_params}")
    with exceptions.raise_on_empty("No updated incident was returned. Please view the logs for more information."):
        return _incidents.update(service_now, sys_id, display_value=display_value, **update_params)[0]


@typechecked
def search(
    service_now: ServiceNow,
    *,
    attributes: Optional[list] = None,
    display_value: Optional[str] = "false",
    **search_params,
) -> List[Incident]:
    """
    Search for Service Now Incidents with multiple properties and values.

    :param service_now: A ServiceNow Adaptor
    :param attributes: A list of attributes to return in the Incident Record.
    :param display_value: Return field display values (true), actual values (false), or both (all).
    :param search_params: Keyword arguments to use in the search. Each keyword must be a valid Incident field.

    :raises exceptions.AuthenticationFailed: if authentication failed using the supplied access token.
    :returns: A list of :ref:`records:Incident` records.
    """
    return _incidents.search(service_now, attributes=attributes, display_value=display_value, **search_params)


@typechecked
def get_element_choices(service_now: ServiceNow, *, element: str, table: str = "incident") -> dict:
    """
    Get valid choices for an Incident element.

    :param service_now: A ServiceNow Adaptor
    :param element: The element to find choices for.
    :param table: Optional. Defaults to `incident`. The table to find the choices in. Incidents inherit
        from tasks, so for some elements you may need to override this value to look in the tasks (or any
        other) table.

    :returns: A choice dictionary where every key is a choice `label`, and every value is the choice `value`.
    """
    return _config_reader._get_element_choices(service_now, table=table, element=element)


@typechecked
def get_valid_states(service_now: ServiceNow) -> dict:
    """
    Get valid Incident states.

    :param service_now: A ServiceNow Adaptor

    :returns: A choice dictionary where every key is a choice `label`, and every value is the choice `value`.
    """
    return get_element_choices(service_now, element="state")


@typechecked
def get_valid_close_codes(service_now: ServiceNow) -> dict:
    """
    Get valid Incident close codes.

    :param service_now: A ServiceNow Adaptor

    :returns: A choice dictionary where every key is a choice `label`, and every value is the choice `value`.
    """
    return get_element_choices(service_now, element="close_code")


@typechecked
def get_valid_priorities(service_now: ServiceNow) -> dict:
    """
    Get valid incident priorities.

    :param service_now: A ServiceNow Adaptor

    :returns: A choice dictionary where every key is a choice `label`, and every value is the choice `value`.
    """
    try:
        return get_element_choices(service_now, element="priority")
    except exceptions.NoSuchObject:
        try:
            return get_element_choices(service_now, element="priority", table="task")
        except exceptions.NoSuchObject:
            raise exceptions.NoSuchObject(
                "Could not find choices for incident priority in either incident table or task table."
            )
