import logging

from typing import List, Optional

from typeguard import typechecked

from ._client import ServiceNow
from . import _requested_items, _config_reader, exceptions
from ._records import RequestedItem


logger = logging.getLogger(__name__)


@typechecked
def get_all(
    service_now: ServiceNow, *, display_value: Optional[str] = "false", attributes: List = None
) -> List[RequestedItem]:
    """
    Get all ServiceNow requested items.

    :param service_now: A ServiceNow Adaptor
    :param attributes: A list of attributes to return in each Requested Item Record.
    :param display_value: Return field display values (true), actual values (false), or both (all).

    :raises exceptions.AuthenticationFailed: if authentication failed using the supplied access token.

    :returns: A list of :ref:`records:Requested Item` records.
    """
    return _requested_items.get_all(service_now, display_value=display_value, attributes=attributes)


@typechecked
def get_by_sys_id(
    service_now: ServiceNow, sys_id: str, *, display_value: Optional[str] = "false", attributes: List = None
) -> RequestedItem:
    """
    Get a ServiceNow requested item by sys_id.

    :param service_now: A ServiceNow Adaptor
    :param sys_id: The sys id of the ServiceNow reqested item.
    :param attributes: A list of attributes to return in the Requested Item Record.
    :param display_value: Return field display values (true), actual values (false), or both (all).

    :raises exceptions.AuthenticationFailed: if authentication failed using the supplied access token.
    :raises exceptions.ServiceNowRequestError: if the request returned an unsuccessful status code.
    :raises exceptions.NoSuchObject: if no matching Requested Item was found.


    :returns: A :ref:`records:Requested Item` record.
    """
    logger.info(f"looking for requested item with sys_id {sys_id}")
    with exceptions.raise_on_empty(f"No requested item found with sys_id {sys_id}"):
        return _requested_items.by_sys_id(service_now, sys_id, display_value=display_value, attributes=attributes)[0]


@typechecked
def get_by_number(
    service_now: ServiceNow, number: str, *, display_value: Optional[str] = "false", attributes: List = None
) -> RequestedItem:
    """
    Get a ServiceNow requested item with a specific number.

    :param service_now: A ServiceNow Adaptor
    :param number: The requested item number to look for. If this is not supplied,
        the number will be established by interacting with the user.
    :param attributes: A list of attributes to return in the Requested Item Record.
    :param display_value: Return field display values (true), actual values (false), or both (all).

    :raises exceptions.AuthenticationFailed: if authentication failed using the supplied access token.
    :raises exceptions.NoSuchObject: if no matching Requested Item was found.

    :returns: A :ref:`records:Requested Item` record.
    """

    logger.info(f"looking for requested item with number {number}")
    with exceptions.raise_on_empty(f"No requested item found with number {number}"):
        return _requested_items.search(service_now, display_value=display_value, attributes=attributes, number=number)[
            0
        ]


@typechecked
def create(service_now: ServiceNow, display_value: Optional[str] = "false", **requested_item_data) -> RequestedItem:
    """
    Create a ServiceNow requested item.

    :param service_now: A ServiceNow Adaptor
    :param requested_item_data: Keyword arguments to use when creating a Requested Item. Each keyword must be a valid
        Requested Item field.
    :param display_value: Return field display values (true), actual values (false), or both (all).

    :raises exceptions.AuthenticationFailed: if authentication failed using the supplied access token.
    :raises exceptions.NoSuchObject: if no updated Requested Item record was returned.

    :returns: A :ref:`records:Requested Item` record.
    """
    logger.info(f"creating requested item with parameters {requested_item_data}")
    with exceptions.raise_on_empty(
        "No updated requested item was returned. Please view the logs for more information."
    ):
        return _requested_items.create(service_now, display_value=display_value, **requested_item_data)[0]


@typechecked
def search(
    service_now: ServiceNow, *, display_value: Optional[str] = "false", attributes: List = None, **search_params
) -> List[RequestedItem]:
    """
    Search for Service Now Requested Items with multiple properties and values.

    :param service_now: A ServiceNow Adaptor
    :param attributes: A list of attributes to return in each Requested Item Record.
    :param search_params: Keyword arguments to use in the search. Each keyword must be a valid Requested Item field.
    :param display_value: Return field display values (true), actual values (false), or both (all).

    :raises exceptions.AuthenticationFailed: if authentication failed using the supplied access token.

    :returns: A list of :ref:`records:Requested Item` records.
    """
    return _requested_items.search(service_now, display_value=display_value, attributes=attributes, **search_params)


@typechecked
def update(service_now: ServiceNow, sys_id, display_value: Optional[str] = "false", **update_params) -> RequestedItem:
    """
    Update a ServiceNow Requested Item.

    :param service_now: A ServiceNow Adaptor.
    :param update_params: Keyword arguments to use when updating a Requested Item. Each keyword must
        be a valid Requested Item field.
    :param display_value: Return field display values (true), actual values (false), or both (all).

    :raises exceptions.AuthenticationFailed: if authentication failed using the supplied access token.
    :raises exceptions.NoSuchObject: if no updated Requested Item record was returned.

    :returns: A :ref:`records:Requested Item` record.
    """
    logger.info(f"updating requested item with sys_id {sys_id} and parameters {update_params}")
    with exceptions.raise_on_empty(
        "No updated requested item was returned. Please view the logs for more information."
    ):
        return _requested_items.update(service_now, sys_id, display_value=display_value, **update_params)[0]


@typechecked
def get_element_choices(service_now: ServiceNow, *, element: str, table: str = "sc_req_item") -> dict:
    """
    Get valid choices for an Requested Item element.

    :param service_now: A ServiceNow Adaptor
    :param element: The element to find choices for.
    :param table: Optional. Defaults to `sc_req_item`. The table to find the choices in. You can override this value
        to look in other tables.

    :returns: A choice dictionary where every key is a choice `label`, and every value is the choice `value`.
    """
    return _config_reader._get_element_choices(service_now, table=table, element=element)


@typechecked
def get_valid_states(service_now: ServiceNow) -> dict:
    """
    Get valid Requested Item states.

    :param service_now: A ServiceNow Adaptor

    :returns: A choice dictionary where every key is a choice `label`, and every value is the choice `value`.
    """
    return get_element_choices(service_now, element="state", table="task")


@typechecked
def get_valid_stages(service_now: ServiceNow) -> dict:
    """
    Get valid Requested Item stages.

    :param service_now: A ServiceNow Adaptor

    :returns: A choice dictionary where every key is a choice `label`, and every value is the choice `value`.
    """
    return get_element_choices(service_now, element="stage")
