"""
Incidents

This module contains actions related to ServiceNow Incidents.
"""
import uuid

from typing import List, Union, Dict, Optional, Any

from service_now import incidents, users, exceptions

from ppa import ui

from ._wrappers import _api_function
from .records import Incident, IncidentData, AccessDetails


INCIDENT_TABLE_HEADER = ["Number", "Short Description", "State", "Assigned To", "Active"]

INCIDENT_TABLE_FIELDS = [
    lambda incident: incident.get("number", ""),
    lambda incident: incident.get("short_description", ""),
    lambda incident: incident.get("state", {}).get("name", ""),
    lambda incident: incident.get("assigned_to", {}).get("name", ""),
    lambda incident: str(incident.get("active")),
]


@_api_function
def search(access_details: AccessDetails, **kwargs: Union[str, int, bool]) -> List[Incident]:
    """Search for incidents. Supports multiple search fields and values.

    Advanced search techniques & filters are not supported yet.

    Each supplied field & value will be converted to _field_=_value_ in the search query.

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __access_details:__ an [AccessDetails Dictionary](../inputs_outputs/#accessdetails)

        - __search_params:__ a dictionary containing [Incident](../inputs_outputs/#incident) fields and values

    ??? success "Output"
        A list of [Incident Dictionaries](../inputs_outputs/#incident).

    ??? example
        Searching for all _active_ incidents.
        ```YAML linenums="1"
        - service_now.incidents.search:
            active: true
          load:
            access_details: service_now_api_info
          save: active_incidents
        ```"""
    return [
        IncidentData(**incident._asdict()).__dict__
        for incident in incidents.search(access_details, display_value="all", **kwargs)
    ]


@_api_function
def get_by_number(access_details: AccessDetails, number: str) -> Incident:
    """Get an incident by its number.

    !!! question "Minimum Plugin Version: 4.0.0"

    ??? info "Input"

        - __access_details:__ an [AccessDetails Dictionary](../inputs_outputs/#accessdetails)

        - __number:__ the incident number to search for

    ??? success "Output"

        A single [Incident Dictionary](../inputs_outputs/#incident).

    !!! warning "Incident Numbers"

        ServiceNow incident numbers are _not unique_ by default, but for simplicity this action assumes they are.

        If you _do_ have shared incident numbers, you can use the [search](#search) action to get multiples.

    ??? example
        ```YAML linenums="1"
        - service_now.incidents.get_by_number:
            number: TEST-INCIDENT
          load:
            access_details: service_now_api_info
          save: test_incident
        ```"""
    return IncidentData(**incidents.get_by_number(access_details, number, display_value="all")[0]._asdict()).__dict__


@_api_function
def update(access_details: AccessDetails, sys_id: str, **update_data: Union[str, int, bool]) -> Incident:
    """Update an incident using one or more sets of fields & values

    !!! question "Minimum Plugin Version: 1.0.0"

        ??? info "Input"

            - __access_details:__ an [AccessDetails Dictionary](../inputs_outputs/#accessdetails)

            - __sys_id:__ the _sys_id_ of the incident to update

            - __update_data:__ a dictionary containing [Incident Dictionary](../inputs_outputs/#incident) fields and values

        ??? success "Output"

            The updated incident as an [Incident Dictionary](../inputs_outputs/#incident).

        ??? example

            Resolving a incident _TEST-INCIDENT_ with a comment and note.

            ```YAML linenums="1"
            - service_now.incidents.update:
                number: TEST-INCIDENT
                state: 6
                close_notes: Resolved via PPA
                comment: Resolving an incident from a PPA task
              load:
                access_details: service_now_api_info
                sys_id: incident.sys_id
              save: resolved_incident
            ```"""
    return IncidentData(**incidents.update(access_details, sys_id, display_value="all", **update_data)._asdict()).__dict__


@_api_function
def get_by_number_interactive(access_details: AccessDetails) -> Incident:
    """Interactively search for an incident using the _number_ field.

    !!! question "Minimum Plugin Version: 1.0.0"

    ??? info "Input"

        - __access_details:__ an [AccessDetails Dictionary](../inputs_outputs/#accessdetails)

    ??? success "Output"
        An [Incident Dictionary](../inputs_outputs/#incident).

    ??? example
        ```YAML linenums="1"
        - service_now.incidents.get_by_number_interactive:
          load:
            access_details: service_now_api_info
          save: incident
        ```"""
    header = ["Number", "Caller Name", "Short Description", "Opened At"]
    input_id = str(uuid.uuid4())
    error_id = str(uuid.uuid4())
    input_table_id = str(uuid.uuid4())
    output_table_id = str(uuid.uuid4())
    progress_id = str(uuid.uuid4())

    def _search():
        while True:
            number = ui.input_text("Incident Number", required=True, element_id=input_id)
            ui.output_progress(f"Looking for Incident {number}...", percent=None, element_id=progress_id)
            try:
                results = incidents.get_by_number(access_details, number, display_value="all")
                if len(results) > 1:
                    ui.output_progress(f"Found {len(results)} Incidents", percent=100, element_id=progress_id)
                else:
                    ui.output_progress(f"Found Incident {number}", percent=100, element_id=progress_id)
                break
            except exceptions.NoSuchObject:
                ui.output_progress("No Incident found", percent=0, element_id=progress_id)
                ui.output_error(f"No Incident found with number {number}. Please try again.", element_id=error_id)
            except Exception as e:
                ui.output_progress("Operation failed", percent=0, element_id=progress_id)
                raise e
        return results

    def _select(results):
        if len(results) == 1:
            return next(iter(results))
        table_data = []
        for incident in results:
            caller = users.get_by_sys_id(access_details, incident.caller_id.value)
            caller_name = getattr(caller, "name", getattr(caller, "user_name"))
            table_data.append([incident.number, caller_name, incident.short_description, incident.opened_at])
        while True:
            indices = ui.input_table("Please Select an Incident", header=header, data=table_data, element_id=input_table_id)
            if any([not indices, len(indices) > 1]):
                ui.output_error("Please select one incident.", element_id=error_id)
                continue
            return next(iter((results[i] for i in indices)))

    def _confirm(incident):
        caller = users.get_by_sys_id(access_details, incident.caller_id.value)
        caller_name = getattr(caller, "name", getattr(caller, "user_name"))
        ui.output_table(
            "Incident Details",
            header=header,
            data=[[incident.number, caller_name, incident.short_description, incident.opened_at]],
            element_id=output_table_id,
        )
        if ui.input_confirm("Is this the correct Incident?", confirm="Yes", cancel="No"):
            return True
        return False

    for _ in range(0, 10):
        results = _search()
        chosen_incident = _select(results)

        if not _confirm(chosen_incident):
            if not ui.input_confirm("No Incident selected, search again?", confirm="Yes", cancel="No"):
                raise exceptions.NoSuchObject("No Incident was selected.")
            ui.output_error("Please search again.")
        else:
            return IncidentData(**chosen_incident._asdict()).__dict__
    raise Exception("No Incident selected in 10 attempts. Operation failed.")


@_api_function
def get_valid_states(access_details: AccessDetails) -> Dict[str, Union[str, int]]:
    """Get valid incident states from the ServiceNow configuration.

    !!! question "Minimum Plugin Version: 3.0.0"

    ??? info "Input"

        - __access_details:__ an [AccessDetails Dictionary](../inputs_outputs/#accessdetails)

    ??? success "Output"

        A dictionary of states in the following format:

        ```YAML linenums="1"
        New: 1
        In Progress: 2
        On Hold: 3
        Resolved: 6
        Closed: 7
        Canceled: 8
        ```

    ??? example

        ```YAML linenums="1"
        - service_now.incidents.get_valid_states:
          load:
            access_details: service_now_api_info
          save: valid_states
        ```"""
    # Cast each value to an int here so it matches the incident class field type.
    return {key: int(value) for key, value in incidents.get_valid_states(access_details).items()}


@_api_function
def get_valid_close_codes(access_details: AccessDetails) -> List[str]:
    """Get valid incident close codes from the ServiceNow configuration.

    !!! question "Minimum Plugin Version: 3.0.0"

    ??? info "Input"

        - __access_details:__ an [AccessDetails Dictionary](../inputs_outputs/#accessdetails)

    ??? success "Output"

        A list of close codes in the following format:

        ```YAML linenums="1"
        - Solved (Work Around)
        - Solved (Permanently)
        - Solved Remotely (Work Around)
        - Solved Remotely (Permanently)
        - Not Solved (Not Reproducible)
        - Not Solved (Too Costly)
        - Closed/Resolved by Caller
        ```

    ??? example

        ```YAML linenums="1"
        - service_now.incidents.get_valid_close_codes:
          load:
            access_details: service_now_api_info
          save: valid_close_codes
        ```"""
    return list(incidents.get_valid_close_codes(access_details).keys())


@_api_function
def get_valid_priorities(access_details: AccessDetails) -> Dict[str, Union[str, int]]:
    """Get valid incident priorities from the ServiceNow configuration.

    !!! question "Minimum Plugin Version: 3.0.0"

    ??? info "Input"

        - __access_details:__ an [AccessDetails Dictionary](../inputs_outputs/#accessdetails)

    ??? success "Output"

        A dictionary of priorities in the following format:

        ```YAML linenums="1"
        1 - Critical: 1
        2 - High: 2
        3 - Moderate: 3
        4 - Low: 4
        5 - Planning: 5
        ```

    ??? example

        ```YAML linenums="1"
        - service_now.incidents.get_valid_priorities:
          load:
            access_details: service_now_api_info
          save: valid_priorities
        ```"""
    # Cast each value to an int here so it matches the incident class field type.
    return {key: int(value) for key, value in incidents.get_valid_priorities(access_details).items()}


@_api_function
def choose_state(access_details: AccessDetails) -> Dict[str, Any]:
    """Get valid incident statuses from ServiceNow & allow the Task Operator to choose one.

   !!! question "Minimum Plugin Version: 3.0.0"

    ??? info "Input"

        - __access_details:__ an [AccessDetails Dictionary](../inputs_outputs/#accessdetails)

    ??? success "Output"

    A dictionary containing the `name` & `value` for the chosen state.

    ??? example

        ```YAML linenums="1"
        - service_now.incidents.choose_state:
          load:
            access_details: service_now_api_info
          save: chosen_state

        - service_now.incidents.update:
          load:
            sys_id: incident.sys_id
            state: chosen_state
            access_details: service_now_api_info
        ```
        """
    valid_states = incidents.get_valid_states(access_details)
    choice = ui.input_choice("Select State", options=list(valid_states.keys()))
    return {"name": choice, "value": valid_states[choice]}


@_api_function
def choose_priority(access_details: AccessDetails) -> Dict[str, Any]:
    """Get valid incident priorities from ServiceNow & allow the Task Operator to choose one.

    !!! question "Minimum Plugin Version: 3.0.0"

    ??? info "Input"

        - __access_details:__ an [AccessDetails Dictionary](../inputs_outputs/#accessdetails)

    ??? success "Output"

    A dictionary containing the `name` & `value` for the chosen priority.

    ??? example

        ```YAML linenums="1"
        - service_now.incidents.choose_priority:
          load:
            access_details: service_now_api_info
          save: chosen_priority

        - service_now.incidents.update:
          load:
            sys_id: incident.sys_id
            priority: chosen_priority
            access_details: service_now_api_info
        ```
    """
    valid_priorities = incidents.get_valid_priorities(access_details)
    choice = ui.input_choice("Select Priority", options=list(valid_priorities.keys()))
    return {"name": choice, "value": valid_priorities[choice]}


@_api_function
def choose_close_code(access_details: AccessDetails) -> str:
    """Get valid incident close codes from ServiceNow & allow the Task Operator to choose one.

    !!! question "Minimum Plugin Version: 2.1.0"

    ??? info "Input"

        - __access_details:__ an [AccessDetails Dictionary](../inputs_outputs/#accessdetails)

    ??? success "Output"

    The chosen close code as a string.

    ??? example

        ```YAML linenums="1"
        - service_now.incidents.choose_close_code:
          load:
            access_details: service_now_api_info
          save: chosen_close_code

        - service_now.incidents.update:
          load:
            sys_id: incident.sys_id
            close_code: chosen_close_code
            access_details: service_now_api_info
        ```"""
    # Close code labels & values are a 1-1 mapping, so we don't
    # need a dictionary here, just the list of keys.
    valid_close_codes = list(incidents.get_valid_close_codes(access_details).keys())
    return ui.input_choice("Select Close Code", options=valid_close_codes)


def output_table(text: str, incident_list: List[Incident]) -> None:
    """Display a list of incidents in a table.

    The table will have the following columns:

    - Number
    - Short Description
    - State
    - Assigned To
    - Active

    !!! question "Minimum Plugin Version: 2.0.0"

    ??? info "Input"

        - __text:__ the title of the table

        - __incident_list:__ a list of [Incident Dictionaries](../inputs_outputs/#incident) to display in the table

    ??? success "Output"

        Nothing is outputted by this action.

    ??? example

        {% raw %}
        ```YAML linenums="1"
        - service_now.users.search:
            user_name: john.smith
          load:
            access_details: service_now_api_info
          save: users

        - ppa_tools.lists.get_first_item:
          load:
            items: users
          save: user

        - service_now.incidents.search:
          load:
            assigned_to: user.sys_id
            access_details: service_now_api_info
          save: user_incidents

        - service_now.incidents.output_table:
            text: Incidents for {{ user.user_name }}
          load:
            incident_list: user_incidents
        ```
        {% endraw %}
    """
    ui.output_table(
        text, header=INCIDENT_TABLE_HEADER, data=[[get(incident) for get in INCIDENT_TABLE_FIELDS] for incident in incident_list],
    )


def input_table(
    text: str, incident_list: List[Incident], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[Incident]:
    """Display a list of incidents in a table, & allow the task operator to make a selection.

    The table will have the following columns:

    - Number
    - Short Description
    - State
    - Assigned To
    - Active

    !!! question "Minimum Plugin Version: 3.0.0"

    ??? info "Input"

        - __text:__ The title of the table

        - __incident_list:__ A list of [Incident Dictionaries](../inputs_outputs/#incident) to display in the table

        - __minimum:__ The minimum number of acceptable selections

        - __maximum:__ The maximum number of acceptable selections

    ??? success "Output"

        A list of [Incident Dictionaries](../inputs_outputs/#incident).

    ??? tip
        - If neither a _minimum_ or _maximum_ is provided, the task operator will be able to submit 0 selections.
        - If _minimum_ or _maximum_ are provided, the operation will repeat until the task operator makes a valid number
            of selections.

    ??? example
        ```YAML linenums="1"
        - service_now.incidents.search:
          load:
            access_details: service_now_api_info
          save: incidents

        - service_now.incidents.input_table:
            text: Please Select 1 Incident
            minimum: 1
            maximum: 1
          load:
            incident_list: incidents
          save: incident
        ```
    """
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    while True:
        indices = ui.input_table(
            text,
            header=INCIDENT_TABLE_HEADER,
            data=[[get(incident) for get in INCIDENT_TABLE_FIELDS] for incident in incident_list],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [incident_list[i] for i in indices]
