from typing import TypedDict, Optional

from distutils.util import strtobool

from datacaster.classes import CastDataClass


def _optional_int(value):
    try:
        return int(getattr(value, "value", None))
    except (ValueError, TypeError):
        return None


def _optional_bool(value):
    return _str_to_bool(getattr(value, "value", None))


def _nested_user(value):
    if value == "":
        return None
    if not all([getattr(value, "display_value"), getattr(value, "value")]):
        return None
    return {"name": getattr(value, "display_value"), "sys_id": getattr(value, "value")}


def _nested_state(value):
    if value == "":
        return None
    if not all([getattr(value, "display_value"), getattr(value, "value")]):
        return None
    return {"name": getattr(value, "display_value"), "value": int(getattr(value, "value"))}


def _nested_link(value):
    if value == "":
        return None
    if not all([getattr(value, "link"), getattr(value, "value")]):
        return None
    return {"link": getattr(value, "link"), "sys_id": getattr(value, "value")}


def _nested_simple_value(named_tuple):
    return getattr(named_tuple, "value", None)


def _str_to_bool(value):
    # strtobool actually returns 0 or 1.
    return bool(strtobool(value)) if all([isinstance(value, str), value]) else None


class UserField(TypedDict):
    name: str
    sys_id: str


class StateField(TypedDict):
    name: str
    value: str


class Link(TypedDict):
    link: str
    value: str


class AccessDetails(TypedDict):
    """
    - Contains ServiceNow API connection details

    - It is used as an _input_ to many of the actions in this plugin

    - See the ___required___ and _optional_ keys below

    ??? danger "Required Keys"

        __address__: ServiceNow instance address.

        __access_token__: An OAuth access token generated for the PPA application in ServiceNow.

    """

    address: str
    access_token: str


class Incident(TypedDict):
    """
    - Contains ServiceNow Incident fields

    ??? info "All Incident Keys"
        - __active__
        - __assigned_to__
        - __assignment_group__
        - __caller_id__
        - __category__
        - __close_code__
        - __close_notes__
        - __closed_at__
        - __closed_by__
        - __comments__
        - __description__
        - __impact__
        - __knowledge__
        - __location__
        - __number__
        - __opened_at__
        - __opened_by__
        - __priority__
        - __resolved_at__
        - __resolved_by__
        - __severity__
        - __short_description__
        - __state__
        - __subcategory__
        - __sys_created_by__
        - __sys_created_on__
        - __sys_id__
        - __sys_updated_by__
        - __urgency__
        - __work_notes__
        - __work_notes_list__"""

    active: bool
    assigned_to: UserField
    assignment_group: str
    caller_id: str
    category: str
    close_code: str
    close_notes: str
    closed_at: str
    closed_by: UserField
    comments: str
    description: str
    impact: int
    knowledge: bool
    location: str
    made_sla: bool
    number: str
    opened_at: str
    opened_by: UserField
    priority: int
    resolved_at: str
    resolved_by: UserField
    severity: str
    short_description: str
    state: StateField
    subcategory: str
    sys_created_by: str
    sys_created_on: str
    sys_id: str
    sys_updated_by: str
    urgency: int
    work_notes: str
    work_notes_list: str


class IncidentData(CastDataClass):
    __annotations__ = Incident.__dict__["__annotations__"]
    __class_config__ = {
        "cast_functions": {
            "types": {
                bool: _optional_bool,
                int: _optional_int,
                str: _nested_simple_value,
                UserField: _nested_user,
                StateField: _nested_state,
            },
        },
    }


class RequestedItem(TypedDict):
    """
    - Contains ServiceNow Requested Item fields

        ??? info "All RequestedItem Keys"

            - __active__
            - __approval__
            - __assigned_to__
            - __cat_item__
            - __close_notes__
            - __closed_at__
            - __closed_by__
            - __comments__
            - __configuration_item__
            - __contact_type__
            - __description__
            - __due_date__
            - __follow_up__
            - __group_list__
            - __impact__
            - __location__
            - __made_sla__
            - __number__
            - __opened_at__
            - __opened_by__
            - __priority__
            - __quantity__
            - __request__
            - __sc_catalog__
            - __short_description__
            - __stage__
            - __state__
            - __sys_created_by__
            - __sys_created_on__
            - __sys_id__
            - __sys_tags__
            - __sys_updated_on__
            - __urgency__
            - __work_end__
            - __work_notes__
    """

    active: bool
    approval: str
    assigned_to: UserField
    cat_item: Link
    close_notes: str
    closed_at: str
    closed_by: UserField
    comments: str
    configuration_item: str
    contact_type: str
    description: str
    due_date: str
    follow_up: str
    group_list: str
    impact: Optional[int]
    location: str
    made_sla: bool
    number: str
    opened_at: str
    opened_by: Link
    priority: int
    quantity: int
    request: Link
    sc_catalog: str
    short_description: str
    stage: str
    state: int
    sys_created_by: str
    sys_created_on: str
    sys_id: str
    sys_tags: str
    sys_updated_on: str
    urgency: int
    work_end: str
    work_notes: str


class RequestedItemData(CastDataClass):
    __annotations__ = RequestedItem.__dict__["__annotations__"]
    __class_config__ = {
        "cast_functions": {
            "types": {bool: _str_to_bool, str: _nested_simple_value, Link: _nested_link, UserField: _nested_user},
            "fields": {"impact": _optional_int, "urgency": _optional_int, "priority": _optional_int},
        },
    }


class User(TypedDict):
    """
    - Contains ServiceNow System User fields

    ??? info "All User Keys"

        - __active__
        - __building__
        - __city__
        - __country__
        - __department__
        - __email__
        - __first_name__
        - __last_login__
        - __last_login_time__
        - __last_name__
        - __location__
        - __locked_out__
        - __manager__
        - __middle_name__
        - __mobile_phone__
        - __name__
        - __phone__
        - __roles__
        - __state__
        - __street__
        - __sys_created_by__
        - __sys_created_on__
        - __sys_id__
        - __sys_tags__
        - __sys_updated_by__
        - __sys_updated_on__
        - __title__
        - __user_name__"""

    active: bool
    building: str
    city: str
    country: str
    department: str
    email: str
    first_name: str
    last_login: str
    last_login_time: str
    last_name: str
    location: str
    locked_out: str
    manager: str
    middle_name: str
    mobile_phone: str
    name: str
    phone: str
    roles: str
    state: str
    street: str
    sys_created_by: str
    sys_created_on: str
    sys_id: str
    sys_tags: str
    sys_updated_by: str
    sys_updated_on: str
    title: str
    user_name: str


class UserData(CastDataClass):
    __annotations__ = User.__dict__["__annotations__"]
    __class_config__ = {
        "cast_functions": {"types": {bool: _str_to_bool, str: _nested_simple_value}},
    }
